/****************************************************************************

        Directory.c

	This file contains the C code that implements the directory
	iteration and file information subsystem.

	This code is intended to be used as a convenient, machine
	independent interface to iterate through the contents of a
	directory.

 ****************************************************************************/

/* Changes:
	getwd changed to GetCWD J. Newmarch Nov 92
	Mar 25 93 - all pointer vbls grounded on declaration
	seekdir(dp) changed to seekdir(dp->filep) J Newmarch Dec 92
	string.h added
	include <unistd.h> added
 */
#include <string.h>
#include <unistd.h>

#include "Directory.h"
#include "RegExp.h"
#include "debug_malloc.h"
#ifdef __QNX__
#include <errno.h>
#endif

extern char *GetCWD (
#ifdef UseFunctionPrototypes
	char *, int
#endif
);
extern int chdir (
#ifdef UseFunctionPrototypes
	const char *
#endif
);
/*--------------------------------------------------------------------------*

        L O W    L E V E L    D I R E C T O R Y    I N T E R F A C E

 *--------------------------------------------------------------------------*/

int DirectoryOpen
#ifdef UseFunctionPrototypes
	(char *dir_name, Directory *dp)
#else
	(dir_name,dp)
char *dir_name;
Directory *dp;

#endif
{
	DirectoryDir(dp) = opendir(dir_name);
	if (DirectoryDir(dp) == NULL) return(FALSE);
	if (DirectoryPathExpand(dir_name,DirectoryPath(dp)) == NULL)
	{
		closedir(DirectoryDir(dp));
		return(FALSE);
	}
	return(TRUE);
} /* End DirectoryOpen */


void DirectoryRestart
#ifdef UseFunctionPrototypes
	(Directory *dp)
#else
	(dp)
Directory *dp;

#endif
{
	rewinddir(DirectoryDir(dp));
} /* End DirectoryRestart */


void DirectoryClose
#ifdef UseFunctionPrototypes
	(Directory *dp)
#else
	(dp)
Directory *dp;

#endif
{
	closedir(DirectoryDir(dp));
} /* End DirectoryClose */


long DirectoryTellPosition
#ifdef UseFunctionPrototypes
	(Directory *dp)
#else
	(dp)
Directory *dp;

#endif
{
#ifndef __QNX__
	return(telldir(DirectoryDir(dp)));
#else
	errno = ENOSYS;
	return -1;
#endif
} /* End DirectoryTellPosition */


void DirectorySetPosition
#ifdef UseFunctionPrototypes
	(Directory *dp, long int pos)
#else
	(dp,pos)
Directory *dp;
long pos;

#endif
{
#ifndef __QNX__
	seekdir(dp -> filep,pos);
#endif
} /* End DirectorySetPosition */


int DirectoryReadNextEntry
#ifdef UseFunctionPrototypes
	(Directory *dp, DirEntry *de)
#else
	(dp,de)
Directory *dp;
DirEntry *de;

#endif
{
	u_short orig_file_type;
	static struct dirent *_ep = NULL;
	static struct stat _lstats,_stats,_dstats;
	char full_path[MAXPATHLEN + 2];
	char temp_path[MAXPATHLEN + 2];

	_ep = readdir(DirectoryDir(dp));
	if (_ep == NULL) return(FALSE);
	strcpy(DirEntryFileName(de),_ep->d_name);
	strcpy(full_path,DirectoryPath(dp));
	strcat(full_path,DirEntryFileName(de));

	if (lstat(full_path,&_lstats) != 0) return(FALSE);

	orig_file_type = _lstats.st_mode & S_IFMT;
	switch (orig_file_type)
	{
	    case S_IFDIR:
		DirEntryType(de) = F_TYPE_DIR;
		break;
	    case S_IFREG:
		DirEntryType(de) = F_TYPE_FILE;
		break;
	    case S_IFCHR:
		DirEntryType(de) = F_TYPE_CHAR_SPECIAL;
		break;
	    case S_IFBLK:
		DirEntryType(de) = F_TYPE_BLOCK_SPECIAL;
		break;
	    case S_IFLNK:
		DirEntryType(de) = F_TYPE_SYM_LINK;
		break;
	    case S_IFSOCK:
		DirEntryType(de) = F_TYPE_SOCKET;
		break;
#ifdef S_IFIFO
	    case S_IFIFO:
		DirEntryType(de) = F_TYPE_FIFO;
		break;
#endif
	    default:
		DirEntryType(de) = orig_file_type;
		break;
	}

	DirEntryIsBrokenLink(de) = FALSE;
	DirEntryIsDirectoryLink(de) = FALSE;
	if (DirEntryIsSymLink(de))			/* Symbolic Link */
	{
		if (stat(full_path,&_stats) != 0)	/* Can't Stat File */
		{
			DirEntryIsBrokenLink(de) = TRUE;
			_stats = _lstats;
		}
		    else				/* Link Not Broken */
		{
#ifdef SLOW_DIRLINK_TEST
			if (DirectoryPathExpand(full_path,temp_path) != NULL)
			{
#else
			if ((_stats.st_mode & S_IFMT) == S_IFDIR)
			{
#endif
				DirEntryIsDirectoryLink(de) = TRUE;
			}

		}
	}
	    else					/* Not Symbolic Link */
	{
		_stats = _lstats;
	}

	FileInfoOrigMode(DirEntrySelfInfo(de)) = _lstats.st_mode;
	FileInfoProt(DirEntrySelfInfo(de)) = _lstats.st_mode & 0777;
	FileInfoUserID(DirEntrySelfInfo(de)) = _lstats.st_uid;
	FileInfoGroupID(DirEntrySelfInfo(de)) = _lstats.st_gid;
	FileInfoFileSize(DirEntrySelfInfo(de)) = _lstats.st_size;
	FileInfoLastAccess(DirEntrySelfInfo(de)) = _lstats.st_atime;
	FileInfoLastModify(DirEntrySelfInfo(de)) = _lstats.st_mtime;
	FileInfoLastStatusChange(DirEntrySelfInfo(de)) = _lstats.st_ctime;
	FileInfoLinkNumber(DirEntrySelfInfo(de)) = _lstats.st_nlink;	/* added JN */

	FileInfoOrigMode(DirEntryActualInfo(de)) = _stats.st_mode;
	FileInfoProt(DirEntryActualInfo(de)) = _stats.st_mode & 0777;
	FileInfoUserID(DirEntryActualInfo(de)) = _stats.st_uid;
	FileInfoGroupID(DirEntryActualInfo(de)) = _stats.st_gid;
	FileInfoFileSize(DirEntryActualInfo(de)) = _stats.st_size;
	FileInfoLastAccess(DirEntryActualInfo(de)) = _stats.st_atime;
	FileInfoLastModify(DirEntryActualInfo(de)) = _stats.st_mtime;
	FileInfoLastStatusChange(DirEntryActualInfo(de)) = _stats.st_ctime;
	FileInfoLinkNumber(DirEntryActualInfo(de)) = _stats.st_nlink;	/* added JN */

	return(TRUE);
} /* End DirectoryReadNextEntry */


char *DirectoryPathExpand
#ifdef UseFunctionPrototypes
	(char *old_path, char *new_path)
#else
	(old_path,new_path)
char *old_path,*new_path;

#endif
{
	register char *p = NULL;
	char path[MAXPATHLEN + 2];

	if (GetCWD(path,MAXPATHLEN) == NULL) return(NULL);
	if (chdir(old_path) != 0) return(NULL);
	if (GetCWD(new_path,MAXPATHLEN) == NULL) strcpy(new_path,old_path);
	if (chdir(path) != 0) return(NULL);
	for (p = new_path; *p != '\0'; p++);
	if ((p != new_path) && *(p - 1) != '/')
	{
		*p++ = '/';
		*p = '\0';
	}
	return(new_path);
} /* End DirectoryPathExpand */


/*---------------------------------------------------------------------------*

             D I R E C T O R Y    E N T R Y    R O U T I N E S

 *---------------------------------------------------------------------------*/

void DirEntryDump
#ifdef UseFunctionPrototypes
	(FILE *fp, DirEntry *de)
#else
	(fp,de)
FILE *fp;
DirEntry *de;

#endif
{
	fprintf(fp,"%20s, Size %7d, Prot %3o\n",
		DirEntryFileName(de),DirEntryFileSize(de),DirEntryProt(de));
} /* End DirEntryDump */
