/*************************************************************************
 * File: builtin.c
 * Author: Jan Newmarch
 * Last modified: $Date: 1993/04/01 04:15:21 $
 * Version: $Revision: 1.11 $
 * Purpose: Handle builtin calls to the file manager. These are calls that
 *	must be done internally (eg change directory) or for efficiency
 *	(none yet)
 * Revision history:
 *	4 Aug 92	filter part added to open dir
 *	6 Aug 92	busy cursor when changing/opening dir
 *      8 Dec 92        added icon pixmap on opendir
 *     18 Dec 92	changed bcopy to memcpy
 *	25 Mar 93	all pointer vbls grounded on definition, many
 *			XtMallocs changed to XtCalloc to trap pointer bugs
 ************************************************************************/ 

/*************************************************************************
 * System includes
 ************************************************************************/ 
#include <stdio.h>

/*************************************************************************
 * Local includes
 ************************************************************************/ 
#include "const.h"
#include "types.h"
#include "xmvararg.h"
#include "xmfm_icon.h"
#include "Panel.h"
#include "PaneB.h"
#include "debug_malloc.h"

/*************************************************************************
 * Extern variables
 ************************************************************************/ 
extern XmString empty_string;
extern int pane_count;

/*************************************************************************
 * Extern functions
 ************************************************************************/ 
#ifndef UseFunctionPrototypes

extern DirectoryMgr *GetFilesInDir ();
extern Widget CreateApplication ();
extern Bool XtChdir ();
extern void BusyCursor ();
extern void UnBusyCursor ();
extern void builtin_info ();
extern void SetActionsInDirsToolbar ();

#else

extern DirectoryMgr *GetFilesInDir (void);
extern Widget CreateApplication (Widget parent);
extern Bool XtChdir (char *dir, Widget w);
extern void BusyCursor (Widget toplevel);
extern void UnBusyCursor (Widget toplevel);
extern void builtin_info (dir_pane_info *dpi);
extern void SetActionsInDirsToolbar (dir_pane_info *dpi);

#endif

/*************************************************************************
 * Functions exported
 ************************************************************************/ 
void builtin_command ();
void builtin_chdir ();
void builtin_opendir ();

/*************************************************************************
 * Variables exported
 ************************************************************************/ 

/*************************************************************************
 * Forward functions
 ************************************************************************/ 

/*************************************************************************
 * Local variables
 ************************************************************************/ 


/*************************************************************************
 * Function: SetWMTitle()
 * Purpose: set the title of window and icon.
 * In parameters: dir, toplevel
 * Precondition:
 * Postcondition:
 ************************************************************************/ 
static void
SetWMTitle
#if UseFunctionPrototypes
	(char *dir, Widget toplevel)
#else
	(dir, toplevel)
	char * dir;
	Widget toplevel;
#endif
{	char *p = NULL;
	int len;

	/* set the new window manager title
	   the icon will only show a small amount, so use end of path
	*/
	len = strlen (dir);
	if (strlen (dir) <= 1)
		/* root */
		p = dir;
	else
	{	/* backup over trailing / in dir */
		p = dir + len - 2;
		while (p > dir && *(p - 1) != '/' )
			p--;
	}

	XtVaSetValues (toplevel,
		XmNtitle, dir,
		XmNiconName, p,
		NULL);
}

/*************************************************************************
 * Function: builtin_chdir ()
 * Purpose: handle a change of directory request
 * In parameters: dir, dpi
 * Precondition: dir is an existing directory relative to current one
 * Postcondition: panes are repainted to show the new directory, and
 *                values are set in dpi for this.
 ************************************************************************/ 
void
builtin_chdir 
#if UseFunctionPrototypes
	(char *dir, dir_pane_info *dpi)
#else
	(dir, dpi)
	char * dir;
	dir_pane_info *dpi;
#endif
{
	DirectoryMgr *dm = NULL;
	Widget *files_toolbar = NULL;
	int i;

#ifdef DEBUG
	fprintf (stderr, "built in: chdir %s\n", dir);
#endif
	if (XtChdir (dir, dpi -> toplevel) == False)
		/* chdir failed, no changes */
		return;

	/* turn off the highlighted dir */
	if (dpi -> file_selected != NULL)
		PaneButtonInvertColours (dpi -> file_button_selected);
	dpi -> file_button_selected = NULL;
	dpi -> file_selected = NULL;

	/* void the file toolbar */
	files_toolbar = dpi -> files_toolbar;
	for (i = 0; i < FILES_TOOLBAR_SIZE; i++)
	{	XtVaSetValues (files_toolbar[i],
				XmNlabelString, (XtArgVal) empty_string,
				NULL);
		XtSetSensitive (files_toolbar[i], False);
	}

	BusyCursor (dpi -> toplevel);
	DirectoryMgrClose (dpi -> directory_manager);
	dm = GetFilesInDir ();
	dpi -> directory_manager = dm;
	PanelSetFiles (dpi);
	UnBusyCursor (dpi -> toplevel);
	SetWMTitle (
		DirectoryPath (DirectoryMgrDir (dpi -> directory_manager)),
		dpi -> toplevel);
}


/*************************************************************************
 * Function: builtin_opendir ()
 * Purpose: handle a change of directory request
 * In parameters: dir, dpi
 * Precondition: dir is an existing directory relative to current one
 * Postcondition: panes are repainted to show the new directory, and
 *                values are set in dpi for this.
 ************************************************************************/ 
void
builtin_opendir
#if UseFunctionPrototypes
	(char *dir, dir_pane_info *dpi)
#else
	(dir, dpi)
	char * dir;
	dir_pane_info *dpi;
#endif
{
	Widget shell = NULL, dir_pane = NULL;
	dir_pane_info *new_dpi = NULL;	
        Pixmap		icon_pixmap; 



#ifdef DEBUG
	fprintf (stderr, "built in: opendir %s\n", dir);
#endif
	if (XtChdir (dir, dpi -> toplevel) == False)
		/* chdir failed, no changes */
		return;
	
	pane_count++;

	shell = XtCreateApplicationShell ("xmfm",
				vendorShellWidgetClass,
				NULL, 0);
/*
	shell = XtAppCreateShell ("xmfm", "XmFm",
				vendorShellWidgetClass, NULL,
                                NULL, 0);
*/
			
        /* set the icon pixmap */

	icon_pixmap = XCreateBitmapFromData(XtDisplay(shell),
			RootWindowOfScreen(XtScreen(shell)),
			xmfm_bits, xmfm_width, xmfm_height);
	XtVaSetValues(shell, XmNiconPixmap, icon_pixmap, NULL);

	dir_pane = CreateApplication (shell);
	XtManageChild (dir_pane);

	XtPopup (shell, XtGrabNone);

        /* set files in panes */
        XtVaGetValues (dir_pane,
                        XmNuserData, (XtArgVal) &new_dpi,
                        NULL);

	/* copy over filters */
	new_dpi -> file_filter = XtMalloc (strlen (dpi -> file_filter) + 1);	
	strcpy (new_dpi -> file_filter, dpi -> file_filter);

	/* a compiled regexp is an opaque string, that I suspect may
	   contain NULL bytes. To stop strcpy stopping on them, use
	   memcpy instead
	*/
	memcpy (new_dpi -> file_filter_regexp, dpi -> file_filter_regexp, 
		FSM_LENGTH);

	new_dpi -> dir_filter = XtMalloc (strlen (dpi -> dir_filter) + 1);	
	strcpy (new_dpi -> dir_filter, dpi -> dir_filter);

	memcpy (new_dpi -> dir_filter_regexp, dpi -> dir_filter_regexp, 
		FSM_LENGTH);

	new_dpi -> executable_filter = 
		XtMalloc (strlen (dpi -> executable_filter) + 1);	
	strcpy (new_dpi -> executable_filter, dpi -> executable_filter);

	memcpy (new_dpi -> executable_filter_regexp,
		dpi -> executable_filter_regexp, 
		FSM_LENGTH);

        PanelSetFiltersInLabels (new_dpi);

	/* and now set the files in the panes */
	BusyCursor (new_dpi -> toplevel);
        PanelSetFiles (new_dpi);
	SetActionsInDirsToolbar (new_dpi);
	UnBusyCursor (new_dpi -> toplevel);

	SetWMTitle (
		DirectoryPath (DirectoryMgrDir (dpi -> directory_manager)),
		dpi -> toplevel);

}



/*************************************************************************
 * Function: builtin_command ()
 * Purpose: handle any of the builtin commands
 * In parameters: command
 * Out parameters:
 * Precondition:command is a non-null builtin command
 * Postcondition: builtin command has been executed
 ************************************************************************/ 
void
builtin_command
#if UseFunctionPrototypes
	(char *command, dir_pane_info *dpi)
#else
	(command, dpi)
	char *command;
	dir_pane_info *dpi;
#endif
{
	/* these start with a $ - lose that, and then look */
	command++;
	
	if (strncmp (command, CHDIR_COMMAND, LEN_CHDIR_COMMAND) == 0)
	{
		/* lose the builtin part */
		command += LEN_CHDIR_COMMAND + 1;
		builtin_chdir (command, dpi);
		return;
	}
	
	if (strncmp (command, OPENDIR_COMMAND, LEN_OPENDIR_COMMAND) == 0)
	{
		/* lose the builtin part */
		command += LEN_OPENDIR_COMMAND + 1;
		builtin_opendir (command, dpi);
		return;
	}
	if (strncmp (command, INFO_COMMAND, LEN_INFO_COMMAND) == 0)
	{
		/* lose the builtin part */
		command += LEN_INFO_COMMAND + 1;
		builtin_info (dpi);
		return;
	}
}
