/****************************************************************************
 * File: refresh.c
 * Author: Jan Newmarch
 * Last modified: $Date: 1993/04/01 04:15:45 $
 * Version: $Revision: 1.6 $
 * Purpose: this file contains routines that look out for a child dying.
 *	This presumably is a function called by executing an action on
 *	a file, and this may have modified the directory contents. The
 *	stuff here catches SIGCHLD events.
 * Revision history:
 *	17 Dec 92	added reference_count checks to dpi
 *	25 Mar 93	all pointer vbls grounded on definition, many
 *			XtMallocs changed to XtCalloc to trap pointer bugs
 ***************************************************************************/

#include "copyright.h"

/***************************************************************************
 * System includes
 ***************************************************************************/
#include <sys/wait.h>
#include <sys/types.h>
#include <signal.h>
#include <sys/stat.h>
#include <stdio.h>

/***************************************************************************
 * Local includes
 ***************************************************************************/
#include "types.h"
#include "DirMgr.h"
#include "const.h"
#include "debug_malloc.h"

/***************************************************************************
 * Extern variables
 ***************************************************************************/

/***************************************************************************
 * Extern functions 
 ***************************************************************************/
extern void ResetPanesAndSelectFile (
#ifdef UseFunctionPrototypes
	dir_pane_info *dpi, char *file_name
#endif
);

/***************************************************************************
 * Functions exported 
 ***************************************************************************/
void child_died (
#ifdef UseFunctionPrototypes
	int
#endif
);
void add_pid (
#ifdef UseFunctionPrototypes
	pid_t pid, dir_pane_info *dpi
#endif
);

/***************************************************************************
 * Variables exported 
 ***************************************************************************/

/***************************************************************************
 * Forward functions
 ***************************************************************************/
static void refresh_dir (
#ifdef UseFunctionPrototypes
	XtPointer client_data, XtIntervalId *id
#endif
);

/***************************************************************************
 * Local variables 
 ***************************************************************************/
typedef struct pid_list_elmt
{
	pid_t pid;
	dir_pane_info *dpi;
	struct pid_list_elmt *next;
}
	pid_pair, *pid_list_t;

static pid_list_t pid_list = NULL;

/***************************************************************************
 * Function: check_dir_update
 * Purpose: set a work-proc if needed to refresh dir contents
 * In parameters: dpi
 * Out parameters:
 * Precondition:
 * Postcondition: work proc set if directory contents have changed
 ***************************************************************************/
/* ARGSUSED */
void
check_dir_update
#ifdef UseFunctionPrototypes
	(dir_pane_info *dpi)
#else
	(dpi)
	dir_pane_info *dpi;
#endif
{	
	DirectoryMgr *dm = NULL;
	DirEntry *de = NULL;
	time_t prev_modified, last_modified;
	struct stat stat_buf;

	(dpi -> reference_count)--;
	/* the pane may have vanished while this process was alive: */
	if (dpi -> being_destroyed)
	{	if (dpi -> reference_count == 0)
			XtFree ((char *) dpi);
		return;
	}

	/* does the dir need updating?
	   first find the info in dpi
	 */
	dm = dpi -> directory_manager;
	DirectoryMgrGotoNamedItem (dm, ".");
	de = DirectoryMgrCurrentEntry (dm);
	prev_modified = DirEntryLastModify (de);

	/* now find current info */
	if (stat (DirectoryPath (DirectoryMgrDir (dm)),
				&stat_buf) == -1)
		return;	/* error - e.g. dir has vanished */
	last_modified = stat_buf.st_mtime;
#ifdef DEBUG_SIGNAL
	fprintf (stderr, "directory previously modified %ld\n\
last modified %ld\n", 
			prev_modified, last_modified);
#endif

	/* add timeout to update if dir has changed */
	if (last_modified > prev_modified)
		XtAddTimeOut (0, refresh_dir, (XtPointer) dpi);
}

/***************************************************************************
 * Function: child_died
 * Purpose: catch a SIGCHLD signal, and set a work-proc if needed
 * In parameters:
 * Out parameters:
 * Precondition:
 * Postcondition: work proc set if directory contents have changed
 ***************************************************************************/
/* ARGSUSED */
void
child_died 
#ifdef UseFunctionPrototypes
	(int sig)
#else
	(sig)
	int sig;
#endif
{	int status;
	pid_t pid;
	pid_list_t p = pid_list,
		   trail = pid_list;

	signal (SIGCHLD, SIG_IGN);

#ifdef DEBUG_SIGNAL
	fprintf (stderr, "SIGCHLD caught. waiting...\n");
#endif
/*
	if ((pid = wait (&status)) == -1)
		return;
*/
	if ((pid = waitpid (0, &status, WNOHANG)) <= 0)
		return;
#ifdef DEBUG_SIGNAL
	fprintf (stderr, "waited\n");
#endif
	signal (SIGCHLD, SIG_IGN);
	while (p != NULL)
	{
		if (p -> pid == pid)
		{
			check_dir_update (p -> dpi);

			/* remove this element anyway */
			if (trail == p)		/* start of list */
				pid_list = p -> next;
			else			/* later in list */
				trail -> next = p -> next;
			XtFree ((char *) p);

			signal (SIGCHLD, child_died );
			return;
		}
		trail = p;
		p = p -> next;

	}
	signal (SIGCHLD, child_died );
}


/***************************************************************************
 * Function: add_pid
 * Purpose: add a pid/dpi pair to the pid list
 * In parameters: pid,dpi
 * Out parameters:
 * Precondition:
 * Postcondition: pid added to list of undead children
 ***************************************************************************/
void
add_pid 
#ifdef UseFunctionPrototypes
	(pid_t pid, dir_pane_info *dpi)
#else
	(pid, dpi)
	pid_t pid;
	dir_pane_info *dpi;

#endif
{
	pid_list_t p;

	p = (pid_list_t) XtCalloc (1, sizeof (struct pid_list_elmt));
	p -> pid = pid;
	p -> dpi = dpi;
	p -> next = pid_list;
	pid_list = p;

	/* update reference count on dpi */
	dpi -> reference_count++;
}


/***************************************************************************
 * Function: refresh_dir
 * Purpose: refresh directory contents
 * In parameters: dpi
 * Out parameters:
 * Precondition:
 * Postcondition: new directory contents showing, focus on slected file
 ***************************************************************************/
/* ARGSUSED */
static void
refresh_dir 
#ifdef UseFunctionPrototypes
	(XtPointer client_data, XtIntervalId *id)
#else
	(client_data, id)
	XtPointer client_data;
	XtIntervalId *id;

#endif
{       char *file_selected = NULL;
	dir_pane_info *dpi = (dir_pane_info *) client_data;

#ifdef DEBUG_SIGNAL
	fprintf (stderr, "Timeout called, refreshing dir\n");
        if (dpi -> file_selected != NULL)
                  fprintf (stderr, "with selected file %s", 
                           dpi -> file_selected);
#endif

	ResetPanesAndSelectFile (dpi, dpi -> file_selected);
	signal (SIGCHLD, child_died);
}
